@@ -16,6 +16,7 @@ from tqdm import tqdm
 from hover import module_config
 from hover.core import Loggable
 from hover.utils.bokeh_helper import auto_label_color
+from hover.utils.misc import current_time
 from bokeh.models import (
     Button,
     Dropdown,
@@ -24,7 +25,11 @@ from bokeh.models import (
     TableColumn,
     HTMLTemplateFormatter,
 )
-from .local_config import dataset_help_widget, COLOR_GLYPH_TEMPLATE
+from .local_config import (
+    dataset_help_widget,
+    COLOR_GLYPH_TEMPLATE,
+    DATASET_SUBSET_FIELD,
+)
 
 
 class SupervisableDataset(Loggable):
@@ -113,15 +118,17 @@ class SupervisableDataset(Loggable):
         self.synchronize_df_to_dictl()
         self.setup_widgets()
         # self.setup_label_coding() # redundant if setup_pop_table() immediately calls this again
+        self.setup_file_export()
         self.setup_pop_table(width_policy="fit", height_policy="fit")
-        self._good("Finished initialization.")
+        self.setup_sel_table(width_policy="fit", height_policy="fit")
+        self._good(f"{self.__class__.__name__}: finished initialization.")
 
     def copy(self, use_df=True):
         """
         ???+ note "Create another instance, copying over the data entries."
             | Param    | Type   | Description                          |
             | :------- | :----- | :----------------------------------- |
-            | `use_df` | `bool` | whether to use the df or dictl form |
+            | `use_df` | `bool` | whether to use the df or dictl form  |
         """
         if use_df:
             self.synchronize_df_to_dictl()
@@ -134,6 +141,44 @@ class SupervisableDataset(Loggable):
             label_key="label",
         )
 
+    def to_pandas(self, use_df=True):
+        """
+        ???+ note "Export to a pandas DataFrame."
+            | Param    | Type   | Description                          |
+            | :------- | :----- | :----------------------------------- |
+            | `use_df` | `bool` | whether to use the df or dictl form  |
+        """
+        if not use_df:
+            self.synchronize_dictl_to_df()
+        dfs = []
+        for _subset in ["raw", "train", "dev", "test"]:
+            _df = self.dfs[_subset].copy()
+            _df[DATASET_SUBSET_FIELD] = _subset
+            dfs.append(_df)
+
+        return pd.concat(dfs, axis=0)
+
+    @classmethod
+    def from_pandas(cls, df, **kwargs):
+        """
+        ???+ note "Import from a pandas DataFrame."
+            | Param    | Type   | Description                          |
+            | :------- | :----- | :----------------------------------- |
+            | `df` | `DataFrame` | with a "SUBSET" field dividing subsets |
+        """
+        dictls = {}
+        for _subset in ["raw", "train", "dev", "test"]:
+            dictls[_subset] = df[df[DATASET_SUBSET_FIELD] == _subset].to_dict(
+                orient="records"
+            )
+        return cls(
+            raw_dictl=dictls["raw"],
+            train_dictl=dictls["train"],
+            dev_dictl=dictls["dev"],
+            test_dictl=dictls["test"],
+            **kwargs,
+        )
+
     def setup_widgets(self):
         """
         ???+ note "Create `bokeh` widgets for interactive data management."
@@ -154,6 +199,12 @@ class SupervisableDataset(Loggable):
             height_policy="fit",
             width_policy="min",
         )
+        self.selection_viewer = Button(
+            label="View Selected",
+            button_type="primary",
+            height_policy="fit",
+            width_policy="min",
+        )
 
         def commit_base_callback():
             """
@@ -198,8 +249,15 @@ class SupervisableDataset(Loggable):
 
         return column(
             self.help_div,
-            row(self.update_pusher, self.data_committer, self.dedup_trigger),
+            row(
+                self.update_pusher,
+                self.data_committer,
+                self.dedup_trigger,
+                self.selection_viewer,
+                self.file_exporter,
+            ),
             self.pop_table,
+            self.sel_table,
         )
 
     def subscribe_update_push(self, explorer, subset_mapping):
@@ -247,6 +305,11 @@ class SupervisableDataset(Loggable):
                     return
 
                 # take selected slice, ignoring ABSTAIN'ed rows
+                # CAUTION: applying selected_idx from explorer.source to self.df
+                #     this assumes that the source and the df have consistent entries.
+                # Consider this:
+                #    keep_cols = self.dfs[sub_k].columns
+                #    sel_slice = explorer.dfs[sub_v].iloc[selected_idx][keep_cols]
                 sel_slice = self.dfs[sub_k].iloc[selected_idx]
                 valid_slice = sel_slice[
                     sel_slice["label"] != module_config.ABSTAIN_DECODED
@@ -277,6 +340,35 @@ class SupervisableDataset(Loggable):
             f"Subscribed {explorer.__class__.__name__} to dataset commits: {subset_mapping}"
         )
 
+    def subscribe_selection_view(self, explorer, subsets):
+        """
+        ???+ note "Enable viewing groups of data entries, specified by a selection in an explorer."
+            | Param            | Type   | Description                            |
+            | :--------------- | :----- | :------------------------------------- |
+            | `explorer`       | `BokehBaseExplorer` | the explorer to register  |
+            | `subsets`        | `list` | subset selections to consider          |
+        """
+        assert (
+            isinstance(subsets, list) and len(subsets) > 0
+        ), "Expected a non-empty list of subsets"
+
+        def callback_view():
+            sel_slices = []
+            for subset in subsets:
+                selected_idx = explorer.sources[subset].selected.indices
+                sub_slice = explorer.dfs[subset].iloc[selected_idx]
+                sel_slices.append(sub_slice)
+
+            selected = pd.concat(sel_slices, axis=0)
+
+            # replace this with an actual display (and analysis)
+            self._callback_update_selection(selected)
+
+        self.selection_viewer.on_click(callback_view)
+        self._good(
+            f"Subscribed {explorer.__class__.__name__} to selection view: {subsets}"
+        )
+
     def setup_label_coding(self, verbose=True, debug=False):
         """
         ???+ note "Auto-determine labels in the dataset, then create encoder/decoder in lexical order."
@@ -328,6 +420,50 @@ class SupervisableDataset(Loggable):
                 if raise_exception:
                     raise ValueError("invalid labels")
 
+    def setup_file_export(self):
+        self.file_exporter = Dropdown(
+            label="Export",
+            button_type="warning",
+            menu=["Excel", "CSV", "JSON", "pickle"],
+            height_policy="fit",
+            width_policy="min",
+        )
+
+        def callback_export(event, path_root=None):
+            """
+            A callback on clicking the 'self.annotator_export' button.
+            Saves the dataframe to a pickle.
+            """
+            export_format = event.item
+
+            # auto-determine the export path root
+            if path_root is None:
+                timestamp = current_time("%Y%m%d%H%M%S")
+                path_root = f"hover-dataset-export-{timestamp}"
+
+            export_df = self.to_pandas(use_df=True)
+
+            if export_format == "Excel":
+                export_path = f"{path_root}.xlsx"
+                export_df.to_excel(export_path, index=False)
+            elif export_format == "CSV":
+                export_path = f"{path_root}.csv"
+                export_df.to_csv(export_path, index=False)
+            elif export_format == "JSON":
+                export_path = f"{path_root}.json"
+                export_df.to_json(export_path, orient="records")
+            elif export_format == "pickle":
+                export_path = f"{path_root}.pkl"
+                export_df.to_pickle(export_path)
+            else:
+                raise ValueError(f"Unexpected export format {export_format}")
+
+            self._good(f"Saved DataFrame to {export_path}")
+
+        # assign the callback, keeping its reference
+        self._callback_export = callback_export
+        self.file_exporter.on_click(self._callback_export)
+
     def setup_pop_table(self, **kwargs):
         """
         ???+ note "Set up a bokeh `DataTable` widget for monitoring subset data populations."
@@ -379,7 +515,7 @@ class SupervisableDataset(Loggable):
             pop_source.data = pop_data
 
             self._good(
-                f"Pop updater: latest population with {len(self.classes)} classes."
+                f"Population updater: latest population with {len(self.classes)} classes."
             )
 
         update_population()
@@ -388,6 +524,37 @@ class SupervisableDataset(Loggable):
         # store the callback so that it can be referenced by other methods
         self._callback_update_population = update_population
 
+    def setup_sel_table(self, **kwargs):
+        """
+        ???+ note "Set up a bokeh `DataTable` widget for viewing selected data points."
+
+            | Param      | Type   | Description                  |
+            | :--------- | :----- | :--------------------------- |
+            | `**kwargs` |        | forwarded to the `DataTable` |
+        """
+        sel_source = ColumnDataSource(dict())
+        sel_columns = [
+            TableColumn(
+                field=_col,
+                title=_col,
+            )
+            for _col in self.dfs["train"].columns
+        ]
+        self.sel_table = DataTable(source=sel_source, columns=sel_columns, **kwargs)
+
+        def update_selection(selected_df):
+            """
+            Callback function.
+            """
+            # push results to bokeh data source
+            sel_source.data = selected_df.to_dict(orient="list")
+
+            self._good(
+                f"Selection updater: latest selection with {selected_df.shape[0]} entries."
+            )
+
+        self._callback_update_selection = update_selection
+
     def df_deduplicate(self):
         """
         ???+ note "Cross-deduplicate data entries by feature between subsets."
