@@ -2,10 +2,13 @@
 ???+ note "Base class(es) for ALL explorer implementations."
 """
 from abc import ABC, abstractmethod
+from collections import OrderedDict, defaultdict
 from bokeh.plotting import figure
-from bokeh.models import ColumnDataSource
+from bokeh.models import ColumnDataSource, Slider
 from hover.core import Loggable
-from .local_config import bokeh_hover_tooltip
+from hover.utils.bokeh_helper import bokeh_hover_tooltip
+from hover.utils.misc import RootUnionFind
+from .local_config import SEARCH_SCORE_FIELD
 
 STANDARD_PLOT_TOOLS = [
     # change the scope
@@ -61,20 +64,20 @@ class BokehBaseExplorer(Loggable, ABC):
         """
         self.figure_kwargs = {
             "tools": STANDARD_PLOT_TOOLS,
-            "tooltips": self._build_tooltip(),
+            "tooltips": self._build_tooltip(kwargs.pop("tooltips", "")),
             # bokeh recommends webgl for scalability
             "output_backend": "webgl",
         }
         self.figure_kwargs.update(kwargs)
+        self.figure = figure(**self.figure_kwargs)
         self.glyph_kwargs = {
             _key: _dict["constant"].copy()
             for _key, _dict in self.__class__.SUBSET_GLYPH_KWARGS.items()
         }
-        self._setup_widgets()
         self._setup_dfs(df_dict)
         self._setup_sources()
+        self._setup_widgets()
         self._activate_search_builtin()
-        self.figure = figure(**self.figure_kwargs)
 
     @classmethod
     def from_dataset(cls, dataset, subset_mapping, *args, **kwargs):
@@ -102,19 +105,26 @@ class BokehBaseExplorer(Loggable, ABC):
 
         return column(self._layout_widgets(), self.figure)
 
-    def _build_tooltip(self):
+    def _build_tooltip(self, extra):
         """
         ???+ note "Define a windowed tooltip which shows inspection details."
+            | Param            | Type   | Description                  |
+            | :--------------- | :----- | :--------------------------- |
+            | `extra`          | `str`  | user-supplied extra HTML |
+
             Note that this is a method rather than a class attribute because
             child classes may involve instance attributes in the tooltip.
         """
-        return bokeh_hover_tooltip(**self.__class__.TOOLTIP_KWARGS)
+        standard = bokeh_hover_tooltip(**self.__class__.TOOLTIP_KWARGS)
+        return f"{standard}\n{extra}"
 
     def _setup_widgets(self):
         """
         ???+ note "High-level function creating widgets for interactive functionality."
         """
         self._info("Setting up widgets")
+        self._dynamic_widgets = OrderedDict()
+        self._dynamic_callbacks = OrderedDict()
         self._setup_search_highlight()
         self._setup_subset_toggle()
 
@@ -158,6 +168,68 @@ class BokehBaseExplorer(Loggable, ABC):
         )
         self.data_key_button_group.on_click(update_data_key_display)
 
+    def value_patch(self, col_original, col_patch, **kwargs):
+        """
+        ???+ note "Allow source values to be dynamically patched through a slider."
+            | Param            | Type   | Description                  |
+            | :--------------- | :----- | :--------------------------- |
+            | `col_original`   | `str`  | column of values before the patch |
+            | `col_patch`      | `str`  | column of list of values to use as patches |
+            | `**kwargs`       |        | forwarded to the slider |
+
+            [Reference](https://github.com/bokeh/bokeh/blob/2.3.0/examples/howto/patch_app.py)
+        """
+        # add a patch slider to widgets, if none exist
+        if "patch_slider" not in self._dynamic_widgets:
+            slider = Slider(start=0, end=1, value=0, step=1, **kwargs)
+            slider.disabled = True
+            self._dynamic_widgets["patch_slider"] = slider
+        else:
+            slider = self._dynamic_widgets["patch_slider"]
+
+        # create a slider-adjusting callback exposed to the outside
+        def adjust_slider():
+            """
+            Infer slider length from the number of patch values.
+            """
+            num_patches = None
+            for _key, _df in self.dfs.items():
+                assert (
+                    col_patch in _df.columns
+                ), f"Subset {_key} expecting column {col_patch} among columns, got {_df.columns}"
+                # find all array lengths; note that the data subset can be empty
+                _num_patches_seen = _df[col_patch].apply(len).values
+                assert (
+                    len(set(_num_patches_seen)) <= 1
+                ), f"Expecting consistent number of patches, got {_num_patches_seen}"
+                _num_patches = _num_patches_seen[0] if _df.shape[0] > 0 else None
+
+                # if a previous subset has implied the number of patches, run a consistency check
+                if num_patches is None:
+                    num_patches = _num_patches
+                else:
+                    assert (
+                        num_patches == _num_patches
+                    ), f"Conflicting number of patches: {num_patches} vs {_num_patches}"
+
+            assert num_patches >= 2, f"Expecting at least 2 patches, got {num_patches}"
+            slider.end = num_patches - 1
+            slider.disabled = False
+
+        self._dynamic_callbacks["adjust_patch_slider"] = adjust_slider
+
+        # create the callback for patching values
+        def update_patch(attr, old, new):
+            for _key, _df in self.dfs.items():
+                # calculate the patch corresponding to slider value
+                _value = [_arr[new] for _arr in _df[col_patch].values]
+                _slice = slice(_df.shape[0])
+                _patch = {col_original: [(_slice, _value)]}
+                self.sources[_key].patch(_patch)
+
+        slider.on_change("value", update_patch)
+        self._good(f"Patching {col_original} using {col_patch}")
+
     def _setup_dfs(self, df_dict, copy=False):
         """
         ???+ note "Check and store DataFrames **by reference by default**."
@@ -204,20 +276,85 @@ class BokehBaseExplorer(Loggable, ABC):
         ???+ note "Create, **(not update)**, `ColumnDataSource` objects."
             Intended to be extended in child classes for pre/post processing.
         """
+        from bokeh.events import SelectionGeometry
+
         self._info("Setting up sources")
         self.sources = {_key: ColumnDataSource(_df) for _key, _df in self.dfs.items()}
         self._postprocess_sources()
 
+        # initialize attributes that couple with sources
+        # extra columns for dynamic plotting
+        self._extra_source_cols = defaultdict(dict)
+        # store the last manual selections
+        self._last_selections = {
+            _key: RootUnionFind(set()) for _key in self.sources.keys()
+        }
+        # store commutative, idempotent index filters
+        self._selection_filters = {
+            _key: RootUnionFind(set()) for _key in self.sources.keys()
+        }
+
+        def store_selection(event):
+            """
+            Keep track of the last manual selection.
+            """
+            # ensure that nothing happens until the selection event is complete
+            if not event.final:
+                return
+
+            # store selection indices
+            for _key, _source in self.sources.items():
+                _selected = _source.selected.indices
+                # use clear() and update() instead of assignment to keep clean references
+                self._last_selections[_key].data.clear()
+                self._last_selections[_key].data.update(_selected)
+
+        def trigger_selection_filters(subsets=None):
+            """
+            Filter selection indices on specified subsets.
+            """
+            if subsets is None:
+                subsets = self.sources.keys()
+            else:
+                assert set(subsets).issubset(
+                    self.sources.keys()
+                ), f"Expected subsets from {self.sources.keys()}"
+
+            for _key in subsets:
+                _selected = self._last_selections[_key].data
+                for _func in self._selection_filters[_key].data:
+                    _selected = _func(_selected, _key)
+                self.sources[_key].selected.indices = list(_selected)
+
+        self._trigger_selection_filters = trigger_selection_filters
+        self.figure.on_event(SelectionGeometry, store_selection)
+        self.figure.on_event(
+            SelectionGeometry,
+            lambda event: self._trigger_selection_filters() if event.final else None,
+        )
+
     def _update_sources(self):
         """
         ???+ note "Update the sources with the corresponding dfs."
-            Note that it seems mandatory to re-activate the search widgets.
-            This is because assigning to `source.data` loses plotting kwargs.
+            Note that the shapes and fields of sources are overriden.
+            Thus supplementary fields (those that do not exist in the dfs),
+            such as dynamic plotting kwargs, need to be re-assigned.
         """
         for _key in self.dfs.keys():
             self.sources[_key].data = self.dfs[_key]
         self._postprocess_sources()
-        self._activate_search_builtin(verbose=False)
+        # self._activate_search_builtin(verbose=False)
+
+        # reset attribute values that couple with sources
+        for _key in self.sources.keys():
+            _num_points = len(self.sources[_key].data["label"])
+            # add extra columns
+            for _col, _fill_value in self._extra_source_cols[_key].items():
+                self.sources[_key].add([_fill_value] * _num_points, _col)
+
+            # clear last selection but keep the set object
+            self._last_selections[_key].data.clear()
+            # DON'T DO: self._last_selections = {_key: set() for _key in self.sources.keys()}
 
     def _postprocess_sources(self):
         """
@@ -238,10 +375,18 @@ class BokehBaseExplorer(Loggable, ABC):
         """
         for _key, _dict in self.__class__.SUBSET_GLYPH_KWARGS.items():
             if _key in self.sources.keys():
+                # determine responding attributes
                 _responding = list(_dict["search"].keys())
+
+                # create a field that holds search results that could be used elsewhere
+                _num_points = len(self.sources[_key].data["label"])
+                self._extra_source_cols[_key][SEARCH_SCORE_FIELD] = 0
+                self.sources[_key].add([0] * _num_points, SEARCH_SCORE_FIELD)
+
+                # make attributes respond to search
                 for _flag, _params in _dict["search"].items():
                     self.glyph_kwargs[_key] = self.activate_search(
-                        self.sources[_key],
+                        _key,
                         self.glyph_kwargs[_key],
                         altered_param=_params,
                     )
@@ -251,13 +396,13 @@ class BokehBaseExplorer(Loggable, ABC):
                     )
 
     @abstractmethod
-    def activate_search(self, source, kwargs, altered_param=("size", 10, 5, 7)):
+    def activate_search(self, subset, kwargs, altered_param=("size", 10, 5, 7)):
         """
         ???+ note "Left to child classes that have a specific feature format."
 
             | Param           | Type    | Description                   |
             | :-------------- | :------ | :---------------------------  |
-            | `source`        | `bool`  | the `ColumnDataSource` to use |
+            | `subset`        | `str`   | the subset to activate search on |
             | `kwargs`        | `bool`  | kwargs for the plot to add to |
             | `altered_param` | `tuple` | (attribute, positive, negative, default) |
         """
@@ -285,8 +430,27 @@ class BokehBaseExplorer(Loggable, ABC):
         self._prelink_check(other)
         # link selection in a bidirectional manner
         sl, sr = self.sources[key], other.sources[other_key]
-        sl.selected.js_link("indices", sr.selected, "indices")
-        sr.selected.js_link("indices", sl.selected, "indices")
+
+        # deprecated: use js_link to sync attributes
+        # sl.selected.js_link("indices", sr.selected, "indices")
+        # sr.selected.js_link("indices", sl.selected, "indices")
+        def left_to_right(attr, old, new):
+            sr.selected.indices = sl.selected.indices[:]
+
+        def right_to_left(attr, old, new):
+            sl.selected.indices = sr.selected.indices[:]
+
+        sl.selected.on_change("indices", left_to_right)
+        sr.selected.on_change("indices", right_to_left)
+
+        # link last manual selections (pointing to the same set)
+        self._last_selections[key].union(other._last_selections[other_key])
+
+        # link selection filter functions (pointing to the same set)
+        self._selection_filters[key].data.update(
+            other._selection_filters[other_key].data
+        )
+        self._selection_filters[key].union(other._selection_filters[other_key])
 
     def link_xy_range(self, other):
         """
