@@ -12,11 +12,11 @@ from .subroutine import (
 )
 from hover.utils.bokeh_helper import servable
 from wasabi import msg as logger
-import pandas as pd
+import numpy as np
 
 
 @servable(title="Snorkel Crosscheck")
-def snorkel_crosscheck(dataset, lf_list, height=600, width=600):
+def snorkel_crosscheck(dataset, lf_list, **kwargs):
     """
     ???+ note "Display the dataset for annotation, cross-checking with labeling functions."
         Use the dev set to check labeling functions; use the labeling functions to hint at potential annotation.
@@ -25,8 +25,7 @@ def snorkel_crosscheck(dataset, lf_list, height=600, width=600):
         | :-------- | :------- | :----------------------------------- |
         | `dataset` | `SupervisableDataset` | the dataset to link to  |
         | `lf_list` | `list`   | a list of callables decorated by `@hover.utils.snorkel_helper.labeling_function` |
-        | `height`  | `int`    | height of each Bokeh explorer plot   |
-        | `width`   | `int`    | width of each Bokeh explorer plot    |
+        | `**kwargs` |       | kwargs to forward to each Bokeh figure |
 
         Expected visual layout:
 
@@ -34,9 +33,17 @@ def snorkel_crosscheck(dataset, lf_list, height=600, width=600):
         | :------------------ | :------------------------- | :----------------- |
         | manage data subsets | inspect labeling functions | make annotations   |
     """
+    layout, _ = _snorkel_crosscheck(dataset, lf_list, **kwargs)
+    return layout
+
+
+def _snorkel_crosscheck(dataset, lf_list, **kwargs):
+    """
+    ???+ note "Cousin of snorkel_crosscheck which exposes objects in the layout."
+    """
     # building-block subroutines
-    snorkel = standard_snorkel(dataset, height=height, width=width)
-    annotator = standard_annotator(dataset, height=height, width=width)
+    snorkel = standard_snorkel(dataset, **kwargs)
+    annotator = standard_annotator(dataset, **kwargs)
 
     # plot labeling functions
     for _lf in lf_list:
@@ -46,14 +53,22 @@ def snorkel_crosscheck(dataset, lf_list, height=600, width=600):
     # link coordinates and selections
     snorkel.link_xy_range(annotator)
     snorkel.link_selection("raw", annotator, "raw")
+    snorkel.link_selection("labeled", annotator, "dev")
 
     sidebar = dataset.view()
     layout = row(sidebar, snorkel.view(), annotator.view())
-    return layout
+
+    objects = {
+        "dataset": dataset,
+        "annotator": annotator,
+        "snorkel": snorkel,
+        "sidebar": sidebar,
+    }
+    return layout, objects
 
 
 @servable(title="Active Learning")
-def active_learning(dataset, vectorizer, vecnet_callback, height=600, width=600):
+def active_learning(dataset, vectorizer, vecnet_callback, **kwargs):
     """
     ???+ note "Display the dataset for annotation, putting a classification model in the loop."
         Currently works most smoothly with `VectorNet`.
@@ -63,8 +78,7 @@ def active_learning(dataset, vectorizer, vecnet_callback, height=600, width=600)
         | `dataset` | `SupervisableDataset` | the dataset to link to  |
         | `vectorizer` | `callable` | the feature -> vector function  |
         | `vecnet_callback` | `callable` | the (dataset, vectorizer) -> `VecNet` function|
-        | `height`  | `int`    | height of each Bokeh explorer plot   |
-        | `width`   | `int`    | width of each Bokeh explorer plot    |
+        | `**kwargs` |       | kwargs to forward to each Bokeh figure |
 
         Expected visual layout:
 
@@ -72,16 +86,31 @@ def active_learning(dataset, vectorizer, vecnet_callback, height=600, width=600)
         | :------------------ | :------------------------ | :----------------- | :------------------ |
         | manage data subsets | inspect model predictions | make annotations   | search -> highlight |
     """
-    # building-block subroutines
-    softlabel = standard_softlabel(dataset, height=height, width=width)
-    annotator = standard_annotator(dataset, height=height, width=width)
-    finder = standard_finder(dataset, height=height, width=width)
+    layout, _ = _active_learning(dataset, vectorizer, vecnet_callback, **kwargs)
+    return layout
 
-    # link coordinates and selections
-    softlabel.link_xy_range(annotator)
-    softlabel.link_xy_range(finder)
-    softlabel.link_selection("raw", annotator, "raw")
-    softlabel.link_selection("raw", finder, "raw")
+
+def _active_learning(dataset, vectorizer, vecnet_callback, **kwargs):
+    """
+    ???+ note "Cousin of active_learning which exposes objects in the layout."
+    """
+    # building-block subroutines
+    softlabel = standard_softlabel(dataset, **kwargs)
+    annotator = standard_annotator(dataset, **kwargs)
+    finder = standard_finder(dataset, **kwargs)
+
+    # link coordinates, omitting the softlabel
+    finder.link_xy_range(annotator)
+
+    # link selections, noting that softlabel does not take "test"
+    for _key in ["raw", "train", "dev"]:
+        softlabel.link_selection(_key, annotator, _key)
+        softlabel.link_selection(_key, finder, _key)
+    finder.link_selection("test", annotator, "test")
+
+    # patch coordinates for representational similarity analysis
+    softlabel.value_patch("x", "x_traj", title="Manifold trajectory step")
+    softlabel.value_patch("y", "y_traj")
 
     # recipe-specific widget
     def setup_model_retrainer():
@@ -98,21 +127,50 @@ def active_learning(dataset, vectorizer, vecnet_callback, height=600, width=600)
             model = vecnet_callback(dataset, vectorizer)
 
             train_loader = dataset.loader("train", vectorizer, smoothing_coeff=0.2)
-            dev_loader = dataset.loader("dev", vectorizer)
+            if dataset.dfs["dev"].shape[0] > 0:
+                dev_loader = dataset.loader("dev", vectorizer)
+            else:
+                dataset._warn("Dev set is empty, borrowing train set for validation.")
+                dev_loader = train_loader
 
             _ = model.train(train_loader, dev_loader, epochs=epochs_slider.value)
             model.save()
             logger.good("-- 1/2: retrained model")
 
-            for _key in ["raw", "train", "dev"]:
-                _probs = model.predict_proba(dataset.dfs[_key]["text"].tolist())
-                _labels = [
-                    dataset.label_decoder[_val] for _val in _probs.argmax(axis=-1)
-                ]
-                _scores = _probs.max(axis=-1).tolist()
-                dataset.dfs[_key]["pred_label"] = pd.Series(_labels)
-                dataset.dfs[_key]["pred_score"] = pd.Series(_scores)
+            # combine inputs and compute outputs of all non-test subsets
+            use_subsets = ("raw", "train", "dev")
+            inps = []
+            for _key in use_subsets:
+                inps.extend(dataset.dfs[_key]["text"].tolist())
 
+            probs = model.predict_proba(inps)
+            labels = [dataset.label_decoder[_val] for _val in probs.argmax(axis=-1)]
+            scores = probs.max(axis=-1).tolist()
+            traj_arr, seq_arr, disparity_arr = model.manifold_trajectory(
+                inps,
+                points_per_step=5,
+            )
+
+            offset = 0
+            for _key in use_subsets:
+                _length = dataset.dfs[_key].shape[0]
+                # skip subset if empty
+                if _length > 0:
+                    _slice = slice(offset, offset + _length)
+                    dataset.dfs[_key]["pred_label"] = labels[_slice]
+                    dataset.dfs[_key]["pred_score"] = scores[_slice]
+                    # for each dimension: all steps, selected slice
+                    _x_traj = traj_arr[:, _slice, 0]
+                    _y_traj = traj_arr[:, _slice, 1]
+                    # for each dimension: selected slice, all steps
+                    _x_traj = list(np.swapaxes(_x_traj, 0, 1))
+                    _y_traj = list(np.swapaxes(_y_traj, 0, 1))
+                    dataset.dfs[_key]["x_traj"] = _x_traj
+                    dataset.dfs[_key]["y_traj"] = _y_traj
+
+                    offset += _length
+
+            softlabel._dynamic_callbacks["adjust_patch_slider"]()
             softlabel._update_sources()
             model_retrainer.disabled = False
             logger.good("-- 2/2: updated predictions. Training button is re-enabled.")
@@ -121,6 +179,16 @@ def active_learning(dataset, vectorizer, vecnet_callback, height=600, width=600)
         return model_retrainer, epochs_slider
 
     model_retrainer, epochs_slider = setup_model_retrainer()
-    sidebar = column(model_retrainer, epochs_slider, dataset.view())
+    sidebar = column(row(model_retrainer, epochs_slider), dataset.view())
     layout = row(sidebar, *[_plot.view() for _plot in [softlabel, annotator, finder]])
-    return layout
+
+    objects = {
+        "dataset": dataset,
+        "annotator": annotator,
+        "finder": finder,
+        "sidebar": sidebar,
+        "softlabel": softlabel,
+        "model_retrainer": model_retrainer,
+        "epochs_slider": epochs_slider,
+    }
+    return layout, objects
