@@ -8,15 +8,14 @@ from .subroutine import standard_annotator, standard_finder
 
 
 @servable(title="Simple Annotator")
-def simple_annotator(dataset, height=600, width=600):
+def simple_annotator(dataset, **kwargs):
     """
     ???+ note "Display the dataset with on a 2D map for annotation."
 
         | Param     | Type     | Description                          |
         | :-------- | :------- | :----------------------------------- |
         | `dataset` | `SupervisableDataset` | the dataset to link to  |
-        | `height`  | `int`    | height of each Bokeh explorer plot   |
-        | `width`   | `int`    | width of each Bokeh explorer plot    |
+        | `**kwargs` |       | kwargs to forward to each Bokeh figure |
 
         Expected visual layout:
 
@@ -24,23 +23,32 @@ def simple_annotator(dataset, height=600, width=600):
         | :------------------ | :----------------- |
         | manage data subsets | make annotations   |
     """
-    annotator = standard_annotator(dataset, height=height, width=width)
-
-    sidebar = dataset.view()
-    layout = row(sidebar, annotator.view())
+    layout, _ = _simple_annotator(dataset, **kwargs)
     return layout
 
 
+def _simple_annotator(dataset, **kwargs):
+    """
+    ???+ note "Cousin of simple_annotator which exposes objects in the layout."
+    """
+    annotator = standard_annotator(dataset, **kwargs)
+
+    sidebar = dataset.view()
+    layout = row(sidebar, annotator.view())
+
+    objects = {"dataset": dataset, "annotator": annotator, "sidebar": sidebar}
+    return layout, objects
+
+
 @servable(title="Linked Annotator")
-def linked_annotator(dataset, height=600, width=600):
+def linked_annotator(dataset, **kwargs):
     """
     ???+ note "Display the dataset on a 2D map in two views, one for search and one for annotation."
 
         | Param     | Type     | Description                          |
         | :-------- | :------- | :----------------------------------- |
         | `dataset` | `SupervisableDataset` | the dataset to link to  |
-        | `height`  | `int`    | height of each Bokeh explorer plot   |
-        | `width`   | `int`    | width of each Bokeh explorer plot    |
+        | `**kwargs` |       | kwargs to forward to each Bokeh figure |
 
         Expected visual layout:
 
@@ -48,13 +56,29 @@ def linked_annotator(dataset, height=600, width=600):
         | :------------------ | :------------------ | :----------------- |
         | manage data subsets | search -> highlight | make annotations   |
     """
-    finder = standard_finder(dataset, height=height, width=width)
-    annotator = standard_annotator(dataset, height=height, width=width)
+    layout, _ = _linked_annotator(dataset, **kwargs)
+    return layout
+
+
+def _linked_annotator(dataset, **kwargs):
+    """
+    ???+ note "Cousin of linked_annotator which exposes objects in the layout."
+    """
+    finder = standard_finder(dataset, **kwargs)
+    annotator = standard_annotator(dataset, **kwargs)
 
     # link coordinates and selections
     finder.link_xy_range(annotator)
-    finder.link_selection("raw", annotator, "raw")
+    for _key in ["raw", "train", "dev", "test"]:
+        finder.link_selection(_key, annotator, _key)
 
     sidebar = dataset.view()
     layout = row(sidebar, finder.view(), annotator.view())
-    return layout
+
+    objects = {
+        "dataset": dataset,
+        "annotator": annotator,
+        "finder": finder,
+        "sidebar": sidebar,
+    }
+    return layout, objects
